// Nathan E. Moore
// April 11, 2004
// Parts of this file are derived from PopupMaker.cpp from the popup2 source.

#include "menusettings.h"
#include "menuitem.h"
#include "menubang.h"
#include "menumessage.h"
#include "folderparser.h"
#include "shellparser.h"
#include "tasksparser.h"
#include "manualparser.h"
#include "dateparser.h"
#include "globals.h"
#include <algorithm>
#include <windows.h>
#include "../nem_utils/ls.h"
#include <cstring>
#include <stdexcept>
#include <shlwapi.h>
#include <strsafe.h>

// default prefix values
static const char G_DEFAULT_PREFIX[] = "Menu";

boost::shared_array<wchar_t> expandWinEvars(boost::shared_array<wchar_t> str);
inline bool hasInsert(char* line);

// This class handles the settable prefix that menu uses for all it's settings
// It works by storing a buffer that looks like "*PREFIX_STR                  "
// and returning a const char* to the correct piece of it.
// This could be copyable but there is no need.
class PrefixedRCString : boost::noncopyable
{
private:
	const static size_t S_DEFAULT_SIZE = 0;
	char* m_str;
	char* m_endprefix;
	size_t m_cursize;

public:
	PrefixedRCString(const char* prefix);
	
	~PrefixedRCString()
	{
		delete m_str;
	}

	const char* getRCStar()
	{
		*m_endprefix = '\0';
		return m_str;
	}

	const char* getRCString(const char* rcstr);
};

// This class handle the creation of MenuSub, ManualParser, and CombiningParser objects
// and the messy logic used to determin when to create them
class MenuBuilder : boost::noncopyable
{
private:
	ManualParserPtr m_manual;
	CombiningParserPtr m_combining;
	std::wstring m_txt;

public:
	MenuBuilder(const std::wstring& txt)
		:m_txt(txt)
	{}

	void appendEntry(MenuEntryPtr& entry, bool column)
	{
		if (!m_manual)
			m_manual = ManualParser::createManualParser(m_txt);
		m_manual->appendEntry(entry, column);
	}

	void appendSeparator(bool column)
	{
		if (!m_manual)
			m_manual = ManualParser::createManualParser(m_txt);
		m_manual->appendSeparator(column);
	}

	void appendParser(ParserPtr& parser, bool inserthere, bool columnson, bool column);
	ParserPtr get();
};

void MenuSettings::configure(Menu& menu)
{
	char rcstr[G_MAX_RC_STRING];
	GetRCString("MenuPrefix", rcstr, G_DEFAULT_PREFIX, G_MAX_RC_STRING);
	PrefixedRCString prefix(rcstr);
	settings(prefix);
	parseStepRC(prefix, menu);
}

void MenuSettings::settings(PrefixedRCString& prefix)
{
	// Popup2 settings that menu can use
	MenuSub::enableHeaders(0 != GetRCBool(prefix.getRCString("NoTitles"), FALSE));
	MenuItem::enableExtensions(0 != GetRCBool(prefix.getRCString("ShowExtension"), TRUE));

	// HotlistName is not prefixed with "Popup" so we check both HotlistName and PrefixHotlistName
	// and use PrefixHotlistName if it there otherwise just Hotlistname
	char hlist[G_MAX_RC_STRING];
	GetRCString(prefix.getRCString("HotlistName"), hlist, G_HOTLIST, G_MAX_RC_STRING);
	if (0 == lstrcmpA(G_HOTLIST, hlist))  // PrefixHotlistName doesn't exsist try HotlistName
		GetRCString("HotlistName", hlist, G_HOTLIST, G_MAX_RC_STRING);

	try
	{
		// can't use the nothrow version here as this is not just some random text, we have a default
		// value to use if conversion fails
		m_hotlist = toUnicode(hlist).get();
	}
	catch (std::exception& e)
	{
		LSLOG(lslog::error << "Using default value for Hotlist name. " << e.what());
		m_hotlist = G_WHOTLIST;
	}

	// Menu settings that Popup2 doesn't use
	char horz[G_MAX_RC_STRING];
	char vert[G_MAX_RC_STRING];
	
	GetRCString(prefix.getRCString("HorzAnimation"), horz, "off", G_MAX_RC_STRING);
	GetRCString(prefix.getRCString("VertAnimation"), vert, "off", G_MAX_RC_STRING);
	MenuSub::setAnimation(horz, vert);

	GetRCString(prefix.getRCString("HorzPosition"), horz, "left", G_MAX_RC_STRING);
	GetRCString(prefix.getRCString("VertPosition"), vert, "top", G_MAX_RC_STRING);
	MenuSub::setPosition(horz, vert);

	MenuSub::enableAutoColumn(0 != GetRCBool(prefix.getRCString("AutoColumn"), TRUE));
}

void MenuSettings::parseStepRC(PrefixedRCString& prefix, Menu& menu)
{
	FILE *pStepRC = LCOpen(NULL);

	char szLineBuffer[MAX_LINE_LENGTH];
	char szFirst[MAX_PATH] = { 0 };
	char szSecond[MAX_PATH] = { 0 };
	char szThird[MAX_PATH] = { 0 };
	char szFourth[MAX_PATH] = { 0 };
	char szFifth[MAX_PATH] = { 0 };

	char *rgszTokens[] = { szFirst, szSecond, szThird, szFourth, szFifth };

	if (!LCReadNextConfig(pStepRC, prefix.getRCStar(), szLineBuffer, MAX_LINE_LENGTH))
		return ; //Reads step.rc line starting with *Popup into szLineBuffer

	int nTokens = LCTokenize(szLineBuffer, rgszTokens, 5, NULL);
	//Tokenizes contents of szLineBuffer into up to 5 tokens,
	//stores them in rgszTokens with anything beyond 5 lost


	// multiple popup mode
	while (nTokens >= 2 && (strcmpi(szSecond, "!New") == 0 || strcmpi(szThird, "!New") == 0))
	{
		std::wstring caption;
		std::string bangcommand;
		char *rootcommand; // if after the bangcommand def there is another bang

		// set hotlist name and bang command acc to relevant arguments
		if (nTokens <= 3)
		{
			caption = m_hotlist;
			bangcommand = szThird;
			rootcommand = NULL;
		}
		else if (nTokens >= 4 && strcmpi(szSecond, "!New") == 0)
		{
			caption = m_hotlist;
			bangcommand = szThird;
			rootcommand = szFourth;
		}
		else if (nTokens >= 5 && strcmpi(szThird, "!New") == 0)
		{
			caption = toUnicodeNoThrow(szSecond).get();
			bangcommand = szFourth;
			rootcommand = szFifth;
		}
		else
		{
			caption = toUnicodeNoThrow(szSecond).get();
			bangcommand = szFourth;
			rootcommand = NULL;
		}

		// sets up a new PopupMenu with the spec'd bang command
		MenuSubPtr newpopup;

		if (rootcommand)
		{
			if (memicmp(rootcommand, "!PopupFolder:", strlen("!PopupFolder:")) == 0)
			{
				rootcommand += strlen("!PopupFolder:");
				newpopup = MenuSub::createMenuSub(buildFolderParser<false>(
					caption, expandWinEvars(toUnicodeNoThrow(rootcommand)).get()), true);
			}
		}
		else
		{
			MenuBuilder builder(caption);
			parseFolder(prefix, pStepRC, builder);
			newpopup = MenuSub::createMenuSub(builder.get(), false);
		}

		menu.addBang(bangcommand, newpopup);

		if (!LCReadNextConfig(pStepRC, prefix.getRCStar(), szLineBuffer, MAX_LINE_LENGTH))
			break;

		szFirst[0] = 0;
		szSecond[0] = 0;
		szThird[0] = 0;
		szFourth[0] = 0;
		szFifth[0] = 0;

		nTokens = LCTokenize(szLineBuffer, rgszTokens, 5, NULL);
	}

	LCClose(pStepRC);
}

void MenuSettings::parseFolder(PrefixedRCString& prefix, FILE *pStepRC, MenuBuilder& builder)
{
	while (true)
	{
		char szLineBuffer[MAX_LINE_LENGTH];

		if (!LCReadNextConfig(pStepRC, prefix.getRCStar(), szLineBuffer, MAX_LINE_LENGTH))
		{
			break;
		}
		try
		{
			if (!parseLine(prefix, pStepRC, builder, szLineBuffer))
				break;
		}
		catch (std::exception& e)
		{
			LSLog(LOG_ERROR, G_MODULE_NAME, e.what());
		}
	}
}

bool MenuSettings::parseLine(PrefixedRCString& prefix, FILE *pStepRC, MenuBuilder& builder, char* pszLine)
{
	// this has the side effect of removing the insert from the string so that the config reader needs no changes.
	bool insert = hasInsert(pszLine);

	char szToken1[MAX_LINE_LENGTH],
		 szToken2[MAX_LINE_LENGTH],
		 szToken3[MAX_LINE_LENGTH],
		 szExtra[MAX_LINE_LENGTH];

	LPSTR lpszTokens[] = { szToken1, szToken2, szToken3 };

	szToken1[0] = szToken2[0] = szToken3[0] = szExtra[0] = '\0';

	int nTokens = LCTokenize(pszLine, lpszTokens, 3, szExtra);

	if (_strnicmp(szToken2, ".icon=", 6) == 0)
	{
		char* rgszTokens[] = { szToken3 };
		char szBuffer[MAX_LINE_LENGTH];

		StringCchCopyA(szToken2, MAX_LINE_LENGTH, szToken3);
		StringCchCopyA(szBuffer, MAX_LINE_LENGTH, szExtra);
		LCTokenize(szBuffer, rgszTokens, 1, szExtra);

		if (!szBuffer[0])
			nTokens--;
	}

	if (nTokens == 3)
	{
		if (!strcmpi(szToken3, "Folder"))
		{
			// Folder
			MenuBuilder tempbuilder(toUnicodeNoThrow(szToken2).get());
			parseFolder(prefix, pStepRC, tempbuilder);
			builder.appendParser(tempbuilder.get(), insert, false, getColumn());
		}
		else if (memicmp(szToken3, "!PopupFolder:", strlen("!PopupFolder:")) == 0)
		{
			// PopupFolder
			LPSTR pszTemp = szToken3 + strlen("!PopupFolder:");
			builder.appendParser(buildFolderParser<false>(toUnicodeNoThrow(szToken2).get(),
				expandWinEvars(toUnicodeNoThrow(pszTemp)).get()), insert, true, getColumn());
		}
		else if (memicmp(szToken3, "!DynamicFolder:", strlen("!DynamicFolder:")) == 0)
		{
			// PopupFolder
			LPSTR pszTemp = szToken3 + strlen("!DynamicFolder:");
			builder.appendParser(buildFolderParser<true>(toUnicodeNoThrow(szToken2).get(),
				expandWinEvars(toUnicodeNoThrow(pszTemp)).get()), insert, true, getColumn());
		}
		else if (memicmp(szToken3, "!PopupDynamicFolder:", strlen("!PopupDynamicFolder:")) == 0)
		{
			// PopupFolder
			LPSTR pszTemp = szToken3 + strlen("!PopupDynamicFolder:");
			builder.appendParser(buildFolderParser<true>(toUnicodeNoThrow(szToken2).get(),
				expandWinEvars(toUnicodeNoThrow(pszTemp)).get()), insert, true, getColumn());
		}
		else if (!strcmpi(szToken2, "!RemoteAmp"))
		{
			ParserPtr tempparser;
			if (PathFileExistsA(szToken3) == TRUE)
				tempparser = createRemoteAmp(G_REMOTEAMP);
			else
				tempparser = createRemoteAmp(toUnicodeNoThrow(szToken3).get());
			builder.appendParser(tempparser, insert, false, getColumn());
		}
		else if (!strcmpi(szToken3, "!DateTime"))
			builder.appendParser(DateParser::createDateParser(toUnicodeNoThrow(szToken2).get()), true, false, getColumn());
		else
		{
			try
			{
				builder.appendParser(makeBangItem(szToken3, toUnicodeNoThrow(szToken2).get()), insert, true, getColumn());
			}
			catch (...)
			{
				MenuEntryPtr entry;
				if (szToken3[0] == '!')
					entry = MenuBang::createMenuBang(toUnicodeNoThrow(szToken2).get(), szToken3, szExtra);
				else
					entry = MenuItem::createMenuItem(toUnicodeNoThrow(szToken2).get(), expandWinEvars(toUnicode(szToken3)).get(), toUnicode(szExtra).get());
				builder.appendEntry(entry, getColumn());
			}
		}
	}
	else if (nTokens == 2)
	{
		if (strcmpi(szToken2, "~Folder") == 0 || strcmpi(szToken2, "~New") == 0)
			return false;
		else if (strcmpi(szToken2, "!Separator") == 0)
			builder.appendSeparator(getColumn());
		else if (strcmpi(szToken2, "!Column") == 0)
			setColumn();
		else
			builder.appendParser(makeBangItem(szToken2), insert, true, getColumn());
	}
	return true;
}

ParserPtr MenuSettings::makeBangItem(const char* item, const std::wstring& title) const
{
	if (strcmpi(item, "!PopupTasks") == 0)
		return TasksParser::createTasksParser(title);
	else if (strcmpi(item, "!PopupControlPanel") == 0)
		return ShellParser::createShellParser(title, CSIDL_CONTROLS);
	else if (strcmpi(item, "!PopupMyComputer") == 0)
		return ShellParser::createShellParser(title, CSIDL_DRIVES);
	else if (strcmpi(item, "!PopupNetwork") == 0)
		return ShellParser::createShellParser(title, CSIDL_NETWORK);
	else if (strcmpi(item, "!PopupPrinters") == 0)
		return ShellParser::createShellParser(title, CSIDL_PRINTERS);
	else if (strcmpi(item, "!PopupRecentDocuments") == 0)
		return ShellParser::createShellParser(title, CSIDL_RECENT);
	else if (strcmpi(item, "!PopupRecycleBin") == 0)
		return ShellParser::createShellParser(title, CSIDL_BITBUCKET);
	else if (strcmpi(item, "!PopupTheDesktop") == 0)
		return ShellParser::createShellParser(title, CSIDL_DESKTOP);
	else if (strcmpi(item, "!PopupConnections") == 0)
		return ShellParser::createShellParser(title, CSIDL_CONNECTIONS);
	else
		throw std::runtime_error(ERR_MSG(std::string(item) + " is not a real item."));
}

ParserPtr MenuSettings::makeBangItem(const char* item) const
{
	if (strcmpi(item, "!PopupTasks") == 0)
		return TasksParser::createTasksParser();
	else if (strcmpi(item, "!PopupControlPanel") == 0)
		return ShellParser::createShellParser(CSIDL_CONTROLS);
	else if (strcmpi(item, "!PopupMyComputer") == 0)
		return ShellParser::createShellParser(CSIDL_DRIVES);
	else if (strcmpi(item, "!PopupNetwork") == 0)
		return ShellParser::createShellParser(CSIDL_NETWORK);
	else if (strcmpi(item, "!PopupPrinters") == 0)
		return ShellParser::createShellParser(CSIDL_PRINTERS);
	else if (strcmpi(item, "!PopupRecentDocuments") == 0)
		return ShellParser::createShellParser(CSIDL_RECENT);
	else if (strcmpi(item, "!PopupRecycleBin") == 0)
		return ShellParser::createShellParser(CSIDL_BITBUCKET);
	else if (strcmpi(item, "!PopupTheDesktop") == 0)
		return ShellParser::createShellParser(CSIDL_DESKTOP);
	else if (strcmpi(item, "!PopupConnections") == 0)
		return ShellParser::createShellParser(CSIDL_CONNECTIONS);
	else
		throw std::runtime_error(ERR_MSG(std::string(item) + " is not a real item."));
}

PrefixedRCString::PrefixedRCString(const char* prefix)
{
	// plus one for the star
	size_t length = strlen(prefix) + 1;
	m_cursize = (length > S_DEFAULT_SIZE) ? (length + S_DEFAULT_SIZE + 1) : S_DEFAULT_SIZE;
	m_str = new char[m_cursize];
	*m_str = '*';
	StringCchCopyA(m_str + 1, m_cursize, prefix);
	m_endprefix = m_str + length;
}

const char* PrefixedRCString::getRCString(const char* rcstr)
{
	size_t length = strlen(rcstr);
	size_t prefixlen = m_endprefix - m_str;
	// resize if needed
	if (length > (m_cursize - prefixlen))
	{
		m_cursize = length + prefixlen + 1;
		char* newstr = new char[m_cursize];
		StringCchCopyA(newstr, m_cursize, m_str);
		delete m_str;
		m_str = newstr;
		m_endprefix = newstr + prefixlen;
	}

	StringCchCopyA(m_endprefix, m_cursize - prefixlen, rcstr);
	// plus one to avoid the star
	return m_str + 1;
}

boost::shared_array<wchar_t> expandWinEvars(boost::shared_array<wchar_t> str)
{
	DWORD needsize = 1 + ExpandEnvironmentStrings(str.get(), NULL, 0);
	boost::shared_array<wchar_t> newstr(new wchar_t[needsize]);
	DWORD good = ExpandEnvironmentStrings(str.get(), newstr.get(), needsize);
	if (0 < good && good <= needsize)
	{
		// semi-colon is the windows default path separator
		std::replace(newstr.get(), newstr.get() + good, L';', G_PATH_SEPARATOR);
		return newstr;
	}
	else
	{
		LSLOG(lslog::error << "Errors encountered expanding the windows' evars");
		return str;
	}
}

ParserPtr createRemoteAmp(const std::wstring& label)
{
	ManualParserPtr remoteamp(ManualParser::createManualParser(label));
	remoteamp->appendEntry(MenuMessage::createMenuMessage(L"Previous", L"Winamp v1.x", std::wstring(), WM_COMMAND, 40044, 0), false);
	remoteamp->appendEntry(MenuMessage::createMenuMessage(L"Play", L"Winamp v1.x", std::wstring(), WM_COMMAND, 40045, 0), false);
	remoteamp->appendEntry(MenuMessage::createMenuMessage(L"Pause", L"Winamp v1.x", std::wstring(), WM_COMMAND, 40046, 0), false);
	remoteamp->appendEntry(MenuMessage::createMenuMessage(L"Stop", L"Winamp v1.x", std::wstring(), WM_COMMAND, 40047, 0), false);
	remoteamp->appendEntry(MenuMessage::createMenuMessage(L"Next", L"Winamp v1.x", std::wstring(), WM_COMMAND, 40048, 0), false);
	return boost::static_pointer_cast<ParserPtr::element_type>(remoteamp);
}

inline bool hasInsert(char* line)
{
	char* insertstart = line + strlen(line) - 6;
	// In a perfect world "insert" would be stored in a constant
	bool rval = 0 == strcmpi(insertstart, "insert");
	if (rval)
		*insertstart = '\0';
	return rval;
}

void MenuBuilder::appendParser(ParserPtr& parser, bool inserthere, bool columnson, bool column)
{
	if (inserthere)
	{
		if (!m_combining)
			m_combining = CombiningParser::createCombiningParser(m_txt);
		if (m_manual)
		{
			m_combining->appendParser(boost::static_pointer_cast<ParserPtr::element_type>(m_manual), false);
			m_manual.reset();
		}
		m_combining->appendParser(parser, column);
	}
	else 
	{
		if (!m_manual)
			m_manual = ManualParser::createManualParser(m_txt);
		m_manual->appendSubmenu(MenuSub::createMenuSub(parser, columnson), column);
	}
}

ParserPtr MenuBuilder::get()
{
	if (m_combining)
	{
		if (m_manual)
			m_combining->appendParser(boost::static_pointer_cast<ParserPtr::element_type>(m_manual), false);
		return boost::static_pointer_cast<ParserPtr::element_type>(m_combining);
	}
	else if (m_manual)
		return boost::static_pointer_cast<ParserPtr::element_type>(m_manual);
	else
		return boost::static_pointer_cast<ParserPtr::element_type>(ManualParser::createManualParser(m_txt));
}
