// Nathan E. Moore
// June 2003

#include "idlist.h"
#include "globals.h"
#include <sstream>
#include <stdexcept>
#include <boost/scoped_array.hpp>
#include <boost/noncopyable.hpp>
#include <shlwapi.h>
#include <atlwin.h>
#include <atluser.h>
#include <crtdbg.h>

namespace util
{
	ItemIdList createVirtualFolder(int csidl)
	{
		ITEMIDLIST* pidl;
		if (FAILED(::SHGetFolderLocation(NULL, csidl, NULL, NULL, &pidl)))
			throw std::runtime_error(ERR_MSG("SHGetFolderLocation Failed"));
		return ItemIdList(pidl);
	}

	CComPtr<IShellFolder> getDesktop()
	{
		CComPtr<IShellFolder> sf;
		if (FAILED(::SHGetDesktopFolder(&(sf.p))))
			throw std::runtime_error(ERR_MSG("SHGetDesktopFolder failed"));
		return sf;
	}

	//
	// ItemIdList
	//
	
	// Determines the size in bytes of the give ITEMIDLIST*
	size_t ItemIdList::sizeOf(ITEMIDLIST* pidl) const
	{
		// a null pointer has no size
		_ASSERT(pidl);

		size_t size = 2;	// start at 2 for the null
		ITEMIDLIST* current = pidl;
		while (current->mkid.cb)
		{
			size += current->mkid.cb;
			current = next(current);
		}
		return size;
	}

	ItemIdList ItemIdList::clone() const
	{
		if (NULL == m_pData)
			return ItemIdList();
		else
		{
			size_t size = sizeOf(m_pData);
			ItemIdList rval;
			rval.AllocateBytes(size);
			::CopyMemory(rval.m_pData, m_pData, size);
			return rval;
		}
	}

	// appends idlist.m_ptr to the end of this.m_ptr
	// for making multilevel pidls
	ItemIdList & ItemIdList::append(const ItemIdList &idlist)
	{
		if (NULL == idlist.m_pData)
			return *this;
		else if (m_pData == NULL)
		{
			Attach(idlist.clone().Detach());
			return *this;
		}
		else
		{
			size_t sizethis = sizeOf(m_pData);
			size_t sizeidlist = sizeOf(idlist.m_pData);
			Reallocate(sizethis + sizeidlist - 2);
			::CopyMemory(reinterpret_cast<unsigned char*>(m_pData) + sizethis - 2, idlist.m_pData, sizeidlist);
			return *this;
		}
	}

	// returns a pointer to the last element of m_list
	ITEMIDLIST* ItemIdList::file() const
	{
		ITEMIDLIST* file = m_pData;
		ITEMIDLIST* end = file;
		while (end->mkid.cb)
		{
			file = end;
			end = next(end);
		}
		return file;
	}

	// returns a new smartptr holding all but the last element in m_ptr
	ItemIdList ItemIdList::getContainingElement() const
	{
		size_t size = sizeOf(m_pData) - sizeOf(file()) + 2;
		ItemIdList newidlist;
		newidlist.AllocateBytes(size);
		CopyMemory(newidlist.m_pData, m_pData, size - 2);
		*(reinterpret_cast<USHORT*>((reinterpret_cast<char*>(newidlist.m_pData) + size))) = 0;
		return newidlist;
	}

	// executes the file represted by the ItemIdList
	void ItemIdList::execute(HWND hwnd)
	{
		wchar_t folder[MAX_PATH];
		::SHGetPathFromIDList(getContainingElement().m_pData, folder);

		SHELLEXECUTEINFO info;
		::ZeroMemory(&info, sizeof(info));
		info.cbSize = sizeof(info);
		info.fMask = SEE_MASK_IDLIST | SEE_MASK_FLAG_NO_UI | SEE_MASK_INVOKEIDLIST;
		info.lpIDList = static_cast<void*>(m_pData);
		info.hwnd = hwnd;
		info.nShow = SW_SHOW;
		info.lpDirectory = folder;

		if (!::ShellExecuteEx(&info))
		{
			std::ostringstream oss;
			oss << "ShellExecuteEx Failed. " << WinErrorMsg(GetLastError()) << " " LOCATION;
			throw std::runtime_error(oss.str());
		}
	}

	std::wstring ItemIdList::getDisplayName() const 
	{
		CComPtr<IShellFolder> parent;
		const ITEMIDLIST* singlelevel;
		if (FAILED(SHBindToParent(m_pData, IID_IShellFolder , reinterpret_cast<void**>(&(parent.p)), &singlelevel)))
			throw std::runtime_error(ERR_MSG("Bind to Parent Failed"));
		STRRET name;
		HRESULT hres = parent->GetDisplayNameOf(singlelevel, SHGDN_NORMAL, &name);
		if (FAILED(hres))
			throw std::runtime_error(ERR_MSG("Could not get display name"));
		wchar_t* strname;
		if (FAILED(StrRetToStr(&name, singlelevel, &strname)))
			throw std::runtime_error(ERR_MSG("StrRetToStr Failed"));
		try
		{
			std::wstring rvalue(strname);
			CoTaskMemFree(strname);
			return rvalue;
		}
		catch (...)
		{
			CoTaskMemFree(strname);
			throw;
		}
	}

	//
	// ContextMenu
	//

	class ContextMenuWindow	:
		public CWindowImpl<ContextMenuWindow>,
		public boost::noncopyable
	{
	private:
		CComPtr<IContextMenu2> m_icm2ptr;

		LRESULT onICM2(UINT msg, WPARAM wparam, LPARAM lparam, BOOL& handled)
		{
			if (m_icm2ptr)
				m_icm2ptr->HandleMenuMsg(msg, wparam, lparam);
			else
				handled = FALSE;
			return 0;
		}

	public:
		explicit ContextMenuWindow(HWND hwnd, const CComPtr<IContextMenu2>& icm2ptr)
				: m_icm2ptr(icm2ptr)
		{
			SubclassWindow(hwnd);
		}

		~ContextMenuWindow()
		{
			UnsubclassWindow();
		}

		BEGIN_MSG_MAP(ContextMenuWindow)
			MESSAGE_HANDLER(WM_INITMENUPOPUP, onICM2)
			MESSAGE_HANDLER(WM_DRAWITEM, onICM2)
			MESSAGE_HANDLER(WM_MENUCHAR, onICM2)
			MESSAGE_HANDLER(WM_MEASUREITEM, onICM2)
		END_MSG_MAP()
	};

	bool popupContextMenu(const ItemIdList &idlist, HwndInterface& hi)
	{
		using std::runtime_error;

		HWND hwnd = hi.getHwnd();

		CComPtr<IShellFolder> folder;
		// this ITEMIDLIST is relative so it needs no clean up
		ITEMIDLIST* filelist;
		if (FAILED(::SHBindToParent(idlist, IID_IShellFolder, reinterpret_cast<void**>(&(folder.p)), const_cast<const ITEMIDLIST**>(&filelist))))
			throw runtime_error(ERR_MSG("SHBindToParent Failed."));

		CComPtr<IContextMenu> cmenu;
		if (FAILED(folder->GetUIObjectOf(hwnd, 1, const_cast<const ITEMIDLIST**>(&filelist), IID_IContextMenu, 0, reinterpret_cast<void**>(&(cmenu.p)))))
			throw runtime_error(ERR_MSG("GetUIObjectOf Failed."));

		WTL::CMenu contextmenu;
		if (!contextmenu.CreatePopupMenu())
			throw runtime_error(ERR_MSG("Unable to create a menu."));
		if (FAILED(cmenu->QueryContextMenu(contextmenu, 0, 1, 0x7FFF, CMF_NORMAL)))
			throw runtime_error(ERR_MSG("QueryContextMenu Failed."));

		POINT pt;
		::GetCursorPos(&pt);
		int cmd = 0;
		// this block is to force the destruction of cmwnd asap
		{
			hi.setMode(HwndInterface::CONTEXT_MENU);
			ContextMenuWindow cmwnd(hwnd, CComQIPtr<IContextMenu2, &IID_IContextMenu2>(cmenu));
			cmd = contextmenu.TrackPopupMenuEx(TPM_RECURSE | TPM_LEFTALIGN | TPM_TOPALIGN | TPM_LEFTBUTTON | TPM_RETURNCMD, pt.x, pt.y, cmwnd);
			hi.setMode(HwndInterface::NORMAL);
		}

		if (cmd != 0)
		{
			::CMINVOKECOMMANDINFOEX cmdinfo;
			::ZeroMemory(&cmdinfo, sizeof(cmdinfo));
			cmdinfo.cbSize = sizeof(cmdinfo);
			cmdinfo.fMask = CMIC_MASK_FLAG_NO_UI | CMIC_MASK_UNICODE;
			cmdinfo.hwnd = hwnd;
			cmdinfo.lpVerb = MAKEINTRESOURCEA(cmd - 1);
			cmdinfo.lpVerbW = MAKEINTRESOURCEW(cmd - 1);
			cmdinfo.nShow = SW_SHOW;

			if (FAILED(cmenu->InvokeCommand(reinterpret_cast<CMINVOKECOMMANDINFO*>(&cmdinfo))))
				throw runtime_error(ERR_MSG("InvokeCommand Failed"));
			else
				return true;
		}
		return false;
	}

	bool popupContextMenu(const std::wstring &filesysobj, HwndInterface& hi)
	{
		// we need to make a copy as ParseDisplayName takes a non const string
		std::wstring::size_type size = filesysobj.size();
		boost::scoped_array<wchar_t> fullpath(new wchar_t[size + 1]);
		filesysobj.copy(fullpath.get(), size);
		*(fullpath.get() + size) = L'\0';

		ITEMIDLIST* pidl;
		if (SUCCEEDED(getDesktop()->ParseDisplayName(hi.getHwnd(), NULL, fullpath.get(), NULL, &pidl, NULL)))
			return popupContextMenu(ItemIdList(pidl), hi);
		else
			return false;
	}
}
