// Nathan E. Moore
// March 22, 2004

#pragma once
#include "globals.h"
#include <boost/shared_ptr.hpp>
#include <boost/noncopyable.hpp>
#include <vector>
#include <utility>
#include <crtdbg.h>

class ParserInterface;
// To cast to one of these typedefs you should do
//		boost::static_pointer_cast<ParserPtr::element_type>()
// as otherwise the typedefs are useless
typedef boost::shared_ptr<ParserInterface> ParserPtr;

class CombiningParser;
typedef boost::shared_ptr<CombiningParser> CombiningParserPtr;

template <bool t_dynamic> class ParserBase;
typedef ParserBase<true> DynamicParserBase;
typedef ParserBase<false> StaticParserBase;

// fwd declaration for HmenuInterface
class HwndInterface;

class NOVTABLE HmenuInterface
{
public:
	// Parsers use these methods to add element to a menu, so their is some control over what they do
	virtual void insertEntry(const MenuEntryPtr& item, unsigned int location, bool column = false) = 0;
	virtual void insertSubmenu(const MenuSubPtr& sub, unsigned int location, bool column = false) = 0;
	virtual void insertSeparator(const unsigned int location, bool column = false) = 0;
	virtual void insertDeadEntry(const std::wstring& txt, unsigned int location, bool column = false) = 0;
	// needed so Parsers can remove items, see the DeleteMenu(HMENU, UINT, UINT) function
	virtual bool deleteMenuElement(unsigned int location) = 0;
	// this is needed so that CombiningParser can start the top of a column.
	virtual void setColumnState(unsigned int location, bool column) = 0;
	virtual bool getColumnState(unsigned int location) = 0;
	virtual ~HmenuInterface()
	{}
};

// This should have another method that would completly reset it, since it's current incarnation
// it allows the implemantation to keep some state info.
class NOVTABLE ParserInterface : public boost::noncopyable
{
public:
	// returns the MenuEntryPtr object for the given id, returned pointer is NULL if not found
	// this method can only becalled between predisplay and postdisplay
	virtual MenuEntryPtr findEntry(unsigned int id) const = 0;
	// works the same as the above method
	virtual MenuSubPtr findSubmenu(HMENU hmenu) const = 0;
	// called before displaying the the given Menu, items should be added starting
	// with the given location, the return value is a offset to the end of this parser's
	// elements
	virtual unsigned int predisplay(HmenuInterface& menu, unsigned int start) = 0;
	// called after the selected menu item was executed, dynamic menu do clean up here.
	// start indicates the location of the first item, returns a count of items remaining
	// in the given menu (Not a count of the number removed).
	virtual unsigned int postdisplay(HmenuInterface& menu, unsigned int start) = 0;
	// does the context menu for this folder
	virtual bool doContextMenu(HwndInterface& hi) = 0;
	// provides the txt used for Headers / FolderEntry. This must be available before the
	// first call to predisplay!
	virtual const std::wstring& getEntryText() const = 0;
	virtual ~ParserInterface()
	{}
};

template <bool t_dynamic>
class ParserBase : public ParserInterface
{
private:
	// needed for lazy creation of menuitems
	bool m_dirty;
	unsigned int m_count;
	std::wstring m_txt;

protected:
	ParserBase()
		:m_dirty(true),
		m_count(0)
	{}

	explicit ParserBase(const std::wstring& txt)
		:m_dirty(true),
		m_count(0),
		m_txt(txt)
	{}

	// called in predisplay if dynamic or dirty, always called at least once
	virtual unsigned int parse(HmenuInterface& menu, unsigned int start) = 0;
	// this method should call endDisplay on any contained submenus.
	// this only needs to do something if you StaticParser contians DynamicParsers
	virtual void cleanupChildren()
	{}
	// actually get rid of all the Menu*Ptrs you are holding, because they no longer
	// exist in the hmenu. This is not absolute, it is just notification that all your
	// items were removed from the menu. This only needs to be overide if t_dyanmic is
	// true, or your parser supports setting the state to dirty (which is not implemented yet)
	// If a StaticParse is a child of another StaticParser and the child never gets displayed
	// this will get called on it because it is still dirty, but you don't need to do anything.
	virtual void clearStorage()
	{}

	// this needs some more work to be implemented.
	//bool isDirty() const
	//{
	//	return m_dirty;
	//}
	//
	//void setDirty()
	//{
	//	m_dirty = true;
	//}

	// only call from the constructor!!
	void setEntryText(const std::wstring& txt)
	{
		m_txt = txt;
	}

public:
	unsigned int predisplay(HmenuInterface& menu, unsigned int start)
	{
		if (m_dirty)
		{
			m_count = parse(menu, start);
			m_dirty = false;
		}
		return m_count;
	}

	unsigned int postdisplay(HmenuInterface& menu, unsigned int start)
	{
		if (t_dynamic || m_dirty)
		{
			for (unsigned int i = 0; i != m_count; ++i)
				menu.deleteMenuElement(start);

			clearStorage();
			m_dirty = true;
			m_count = 0;
			return 0;
		}
		else
			cleanupChildren();
		return m_count;
	}

	const std::wstring& getEntryText() const
	{
		return m_txt;
	}
};

// this should be moved to it's own file.
class CombiningParser : public ParserInterface
{
private:
	// bool value stores columning info.
	typedef std::vector<std::pair<ParserPtr, bool> > ParserVector;
	ParserVector m_parsers;
	std::wstring m_txt;

	explicit CombiningParser(const std::wstring& txt)
		: m_txt(txt)
	{}

public:
	void appendParser(const ParserPtr& parser, bool column)
	{
		m_parsers.push_back(std::make_pair(parser, column));
	}

	// ParserInterface
	MenuEntryPtr findEntry(unsigned int id) const
	{
		for (ParserVector::const_iterator i = m_parsers.begin(); i != m_parsers.end(); ++i)
		{
			MenuEntryPtr entry = (*i).first->findEntry(id);
			if (entry)
				return entry;
		}

		return MenuEntryPtr();
	}

	MenuSubPtr findSubmenu(HMENU hmenu) const
	{
		for (ParserVector::const_iterator i = m_parsers.begin(); i != m_parsers.end(); ++i)
		{
			MenuSubPtr sub = (*i).first->findSubmenu(hmenu);
			if (sub)
				return sub;
		}

		return MenuSubPtr();
	}

	unsigned int predisplay(HmenuInterface& menu, unsigned int start)
	{
		unsigned int location = start;
		// we need to carry the previous parser's column state to the next iteration if
		// that parser doesn't add any items.
		bool carryovercolumn = false;
		for (ParserVector::const_iterator i = m_parsers.begin(); i != m_parsers.end(); ++i)
		{
			unsigned int temp = (*i).first->predisplay(menu, location);
			if (temp)
			{
				// we don't want to turn the column off it is already on
				menu.setColumnState(location, carryovercolumn || (*i).second || menu.getColumnState(location));
				carryovercolumn = false;
			}
			else
				carryovercolumn = (*i).second;

			location += temp;
		}

		return location - start;
	}

	unsigned int postdisplay(HmenuInterface& menu, unsigned int start)
	{
		unsigned int location = 0;
		for (ParserVector::const_iterator i = m_parsers.begin(); i != m_parsers.end(); ++i)
			location += (*i).first->postdisplay(menu, location);

		return location - start;
	}

	bool doContextMenu(HwndInterface&)
	{
		return false;
	}

	const std::wstring& getEntryText() const
	{
		return m_txt;
	}

	static CombiningParserPtr createCombiningParser(const std::wstring& txt)
	{
		CombiningParserPtr temp(new CombiningParser(txt));
		return temp;
	}
};